# -*- coding: utf-8 -*-
"""
    pygments.lexers.ki
    ~~~~~~~~~~~~~~~~~~~~

    Lexers for the Ki language.

    :copyright: Copyright 2015-2016 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, include, bygroups, words, default
from pygments.token import Text, Comment, Operator, Keyword, Name, String, \
                           Number, Punctuation, Whitespace

__all__ = ['KiLexer']


class KiLexer(RegexLexer):
    """
    Lexer for the Ki programming language

    .. versionadded:: 2.0.2
    """
    name = 'Ki'
    filenames = ['*.ki']
    aliases = ['ki']
    mimetypes = ['text/ki']

    tokens = {
        'root': [
            # Whitespace and Comments
            (r'\n', Whitespace),
            (r'\s+', Whitespace),
            (r'#.*$', Comment.Single),

            # Importing
            (r'(from)((?:\s|\\\s)+)', bygroups(Keyword.Namespace, Text),
             'fromimport'),
            (r'(import)((?:\s|\\\s)+)', bygroups(Keyword.Namespace, Text),
             'import'),

            # Keywords
            (words((
                'as', 'case', 'const', 'else', 'enum', 'fail', 'fallthrough',
                'flag', 'fn', 'fntype', 'for', 'gen', 'get', 'if', 'iferr',
                'in', 'loop', 'pred', 'range', 'return', 'static', 'struct',
                'switch', 'trait', 'type', 'while', 'with', 'yield'),
               suffix=r'\b'),
             Keyword),
            (r'mod\b', Keyword.Namespace),
            (r'(true|false)\b', Keyword.Constant),
            (r'(var|val|sharedvar|sharedval|weakvar|weakval|refvar|refval)\b',
             Keyword.Declaration),
            (words(('bool', 'int', 'uint', 's8', 's16', 's32', 's64', 'u8',
                    'u16', 'u32', 'u64', 'flt', 'f32', 'f64', 'dec', 'rune',
                    'str', 'String', 'array', 'dynarray', 'error'),
                   suffix=r'\b'),
             Keyword.Type),
            (words(('print', 'echo', 'die',), suffix=r'\b'), Name.Builtin),

            # Labels
            (r'(break|continue)(\s*)(\'[A-Za-z_]\w*)?',
             bygroups(Keyword, Text.Whitespace, Name.Label)),

            # Character Literal
            (r"""'(\\['"\\nrt]|\\x[0-7][0-9a-fA-F]|\\0"""
             r"""|\\u\{[0-9a-fA-F]{1,6}\}|.)'""",
             String.Char),

            # Binary Literal
            (r'0b[01_]+', Number.Bin, 'number_lit'),

            # Octal Literal
            (r'0o[0-7_]+', Number.Oct, 'number_lit'),

            # Hexadecimal Literal
            (r'0[xX][0-9a-fA-F_]+', Number.Hex, 'number_lit'),

            # Decimal Literal
            (r'[0-9][0-9_]*(\.[0-9_]+[eE][+\-]?[0-9_]+|'
             r'\.[0-9_]*(?!\.)|[eE][+\-]?[0-9_]+)', Number.Float, 'number_lit'),
            (r'[0-9][0-9_]*', Number.Integer, 'number_lit'),

            # String Literal
            (r'b"', String, 'bytestring'),
            (r'"', String, 'string'),
            (r'b?r(#*)".*?"\1', String),

            # Operators and Punctuation
            (r'[{}()\[\]:,.]', Punctuation),
            (r'[+\-*/%&|<>^!~=]', Operator),

            # Identifier
            (r'[a-zA-Z_]\w*', Name),
            include('meta')
        ],
        'import': [
            (r'(?:[ \t]|\\\n)+', Text),
            (r'as\b', Keyword.Namespace),
            (r',', Operator),
            (r'[a-zA-Z_][\w.]*', Name.Namespace),
            default('#pop')  # all else: go back
        ],
        'fromimport': [
            (r'(?:[ \t]|\\\n)+', Text),
            (r'import\b', Keyword.Namespace, '#pop'),
            # if None occurs here, it's "raise x from None", since None can
            # never be a module name
            (r'None\b', Name.Builtin.Pseudo, '#pop'),
            # sadly, in "raise x from y" y will be highlighted as namespace too
            (r'[a-zA-Z_.][\w.]*', Name.Namespace),
            # anything else here also means "raise x from y" and is therefore
            # not an error
            default('#pop'),
        ],
        'number_lit': [
            (r'[us](8|16|32|64|size)*', Keyword, '#pop'),
            (r'f(32|64)', Keyword, '#pop'),
            (r'd', Keyword, '#pop'),
            default('#pop'),
        ],
        'string': [
            (r'"', String, '#pop'),
            (r"""\\['"\\nrt]|\\x[0-7][0-9a-fA-F]|\\0"""
             r"""|\\u\{[0-9a-fA-F]{1,6}\}""", String.Escape),
            (r'[^\\"]+', String),
            (r'\\', String),
        ],
        'bytestring': [
            (r"""\\x[89a-fA-F][0-9a-fA-F]""", String.Escape),
            include('string'),
        ],
        'meta': [
            (r'@\w+((:|\.)\w+)*', Name.Other),
            ('[a-zA-Z_]\w*', Name),
        ],
    }

